#include "mpi.h"
#include "headers.h"

#include <stdlib.h>
#include <string.h>
#include <math.h>
/**************************************************************/
/* conjugate gradient iteration                               */
/**************************************************************/
void cg(int rank,
        int left_neighbour,int right_neighbour,
        int lower_neighbour,int upper_neighbour,
        int x_int_loc,int  y_int_loc,
        int maxit, double eps, double damp,
        int dof_loc, int *KCol, int *RowPtr,
        double *matrix_entries, double *sol, double *rhs,
        double *defect)
{
  int i, k;
  double *r, *d, residual, tau, d0, d1, beta;

  // allocate arrays
  if (!(r=malloc(dof_loc*sizeof(double))))
     MPI_Abort(MPI_COMM_WORLD, 1);
  if (!(d=malloc(dof_loc*sizeof(double))))
     MPI_Abort(MPI_COMM_WORLD, 1);

  k = 0;
  /* compute defect, stored inconsistently */
  MatVecDefect(dof_loc, RowPtr, KCol, matrix_entries, sol, 
               rhs, r);
  for (i=0;i<dof_loc;i++)
     r[i] = -r[i];

  /* copy r to defect */
  memcpy(defect, r, dof_loc*sizeof(double));

  /* make defect consistent */
  MakeVectorConsistent(left_neighbour, right_neighbour,
                       lower_neighbour, upper_neighbour,
                       x_int_loc, y_int_loc, defect);
  /* compute inner product of r and defect, residual is stored
     inconsistently */
  residual = 0;
  for (i=0;i<dof_loc;i++)
     residual +=  r[i] * defect[i];
  /* make residual consistent */
  MPI_Allreduce(&residual, &d0, 1, MPI_DOUBLE, MPI_SUM, MPI_COMM_WORLD);

  /* stopping criterion fulfilled */
  if (sqrt(d0) < eps)
     return;

  /*if (rank==0)
    printf("ite %3d %8.6e\n",k, d0);*/

  for (i=0;i<dof_loc;i++)
     d[i] = -defect[i];

  /* loop */
  while(1)
  {
     k++;
     /*  r= Ad, r stored inconsistently */
     MatVec(dof_loc, RowPtr, KCol, matrix_entries, d, defect); 
     /* (defect,d) */
     residual = 0;
     for (i=0;i<dof_loc;i++)
        residual += defect[i] * d[i];
     /* make inner product consistently */
     MPI_Allreduce(&residual, &tau, 1, MPI_DOUBLE, MPI_SUM, 
                   MPI_COMM_WORLD);
     tau = d0/tau; 
     /* compute updates, r is after this stored consistently */
     for (i=0;i<dof_loc;i++)
     {
        sol[i] += tau * d[i];
        r[i] += tau * defect[i];
        defect[i] = r[i];
     }
     /* make defect consistent */
     MakeVectorConsistent(left_neighbour, right_neighbour,
                          lower_neighbour, upper_neighbour,
                          x_int_loc, y_int_loc, defect);
     residual = 0;
     for (i=0;i<dof_loc;i++)
        residual +=  r[i] * defect[i];  
     MPI_Allreduce(&residual, &d1, 1, MPI_DOUBLE, MPI_SUM, 
                   MPI_COMM_WORLD);
     /*if (rank==0)
       printf("ite %3d %8.6e rate %g\n",k, sqrt(d1), sqrt(d1)/sqrt(d0));*/
     /* stopping criterion fulfilled */
     if (sqrt(d1) < eps)
        break;
     if (k>maxit)
        break;
     beta = d1/d0;  
     d0 = d1;
     /* update d */
     for (i=0;i<dof_loc;i++)
        d[i] = -defect[i] + beta * d[i];
  }

  if (rank==0)
    printf("CG: iterations %d residual %g\n", k, sqrt(d1));

  free(r);
  free(d);
  
}
